@@ -660,6 +660,25 @@ void KArchiveTest::testTarDirectoryTwice() // bug 206994
 
     QCOMPARE(listing.count(), 3);
 }
+
+void KArchiveTest::testTarIgnoreRelativePathOutsideArchive()
+{
+    // This test extracts a Tar archive that contains a relative path "../foo" pointing
+    // outside of the archive directory. For security reasons extractions should only
+    // be allowed within the extracted directory as long as not specifically asked.
+
+    KTar tar(QString::fromLatin1(KDESRCDIR) + QLatin1String("tar_relative_path_outside_archive.tar.bz2"));
+    QVERIFY(tar.open(QIODevice::ReadOnly));
+
+    const KArchiveDirectory *dir = tar.directory();
+    KTempDir tmpDir;
+    const QString dirName = tmpDir.name();
+
+    dir->copyTo(dirName);
+    qDebug() << (dirName + "../foo") << QFile::exists(dirName + "../foo");
+    QVERIFY(!QFile::exists(dirName + "../foo"));
+    QVERIFY(QFile::exists(dirName + "/foo"));
+}
 ///
 
 static const char s_zipFileName[] = "karchivetest.zip";
