@@ -1,6 +1,6 @@
 /*
    This file is part of the Nepomuk KDE project.
-   Copyright (C) 2008-2010 Sebastian Trueg <trueg@kde.org>
+   Copyright (C) 2008-2012 Sebastian Trueg <trueg@kde.org>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -31,8 +31,12 @@
 #include "negationterm.h"
 #include "comparisonterm.h"
 #include "resourcetypeterm.h"
+#include "resourcetypeterm_p.h"
+#include "optionalterm.h"
 #include "queryserializer.h"
 #include "queryparser.h"
+#include "util.h"
+#include "filequery.h"
 
 #include <QtCore/QDebug>
 #include <QtCore/QDateTime>
@@ -46,6 +50,7 @@
 #include <Soprano/Vocabulary/RDFS>
 #include <Soprano/Vocabulary/RDF>
 #include <Soprano/Vocabulary/NRL>
+#include <Soprano/Vocabulary/NAO>
 
 #include "resourcemanager.h"
 #include "literal.h"
@@ -65,70 +70,124 @@ WHERE { ?s ?p ?o .
 ORDER BY  desc (?
 */
 
-namespace {
-    QStringList urlsToStringList( const KUrl::List& urls, KUrl::AdjustPathOption opt ) {
-        QStringList lst;
-        for( KUrl::List::ConstIterator it = urls.constBegin();
-             it != urls.constEnd(); ++it ) {
-            lst.append( it->url(opt) );
+
+Nepomuk::Query::Term Nepomuk::Query::QueryPrivate::createFolderFilter() const
+{
+    Term mainTerm;
+
+    if ( !m_includeFolders.isEmpty() ) {
+        QStringList includeFilter;
+        for( QHash<KUrl, bool>::ConstIterator it = m_includeFolders.constBegin();
+             it != m_includeFolders.constEnd(); ++it ) {
+            const QString urlStr = it.key().url(KUrl::AddTrailingSlash);
+            if( it.value() )
+                includeFilter.append( QString::fromLatin1("(^%1)").arg( urlStr ) );
+            else
+                includeFilter.append( QString::fromLatin1("(^%1[^/]*$)").arg( urlStr ) );
         }
-        return lst;
+        mainTerm = mainTerm && ComparisonTerm(
+                    Nepomuk::Vocabulary::NIE::url(),
+                    LiteralTerm(includeFilter.join( "|" )),
+                    ComparisonTerm::Regexp);
     }
+
+    if ( !m_excludeFolders.isEmpty() ) {
+        mainTerm = mainTerm && NegationTerm::negateTerm(
+                    ComparisonTerm(
+                        Nepomuk::Vocabulary::NIE::url(),
+                        LiteralTerm(QString::fromLatin1("^(%1)").arg( m_excludeFolders.toStringList(KUrl::AddTrailingSlash).join( "|" ) )),
+                        ComparisonTerm::Regexp));
+    }
+    return mainTerm;
 }
 
 
-QString Nepomuk::Query::QueryPrivate::createFolderFilter( const QString& resourceVarName, QueryBuilderData* qbd ) const
+Nepomuk::Query::Term Nepomuk::Query::QueryPrivate::optimizeEvenMore(const Nepomuk::Query::Term& term) const
 {
-    if ( m_includeFolders.count() + m_excludeFolders.count() ) {
-        QString uriVarName = qbd->uniqueVarName();
-        QString filter = resourceVarName + ' ' + Soprano::Node::resourceToN3( Nepomuk::Vocabulary::NIE::url() ) + ' ' + uriVarName + QLatin1String( " . " );
-        if ( !m_includeFolders.isEmpty() ) {
-            filter += QString( " FILTER(REGEX(STR(%1), \"^(%2)\")) ." ).arg( uriVarName ).arg( urlsToStringList(m_includeFolders, KUrl::AddTrailingSlash).join( "|" ) );
+    using namespace Nepomuk::Query;
+
+    //
+    // Merge ResourceTypeTerms which are combined in an OrTerm
+    //
+    if(term.type() == Term::Or) {
+        QList<Term> subTerms = term.toOrTerm().subTerms();
+        QList<ResourceTypeTerm> typeTerms;
+        QList<Term>::iterator it = subTerms.begin();
+        while(it != subTerms.end()) {
+            if(it->type() == Term::ResourceType) {
+                typeTerms << it->toResourceTypeTerm();
+                it = subTerms.erase(it);
+            }
+            else {
+                ++it;
+            }
         }
-        if ( !m_excludeFolders.isEmpty() ) {
-            filter += QString( " FILTER(!REGEX(STR(%1), \"^(%2)\")) ." ).arg( uriVarName ).arg( urlsToStringList(m_excludeFolders,  KUrl::AddTrailingSlash).join( "|" ) );
+
+        if(typeTerms.count() > 1) {
+            ResourceTypeTerm newTypeTerm;
+            foreach(const ResourceTypeTerm& rtt, typeTerms) {
+                static_cast<ResourceTypeTermPrivate*>(newTypeTerm.d_ptr.data())->m_types << rtt.type();
+            }
+            subTerms << newTypeTerm;
         }
-        return filter;
+        else if(typeTerms.count() == 1) {
+            subTerms += typeTerms.first();
+        }
+
+        if(subTerms.count() > 1)
+            return OrTerm(subTerms);
+        else if(subTerms.count() == 1)
+            return subTerms.first();
+        else
+            return OrTerm();
     }
+
+    //
+    // For AndTerms we simply need to optimize the subterms
+    //
+    else if(term.type() == Term::And) {
+        AndTerm newAndTerm;
+        //
+        // Another small optimization: we sort all the ComparisonTerms which have a variable name set
+        // at the beginning. This allows to force this variable name in other ComparisonTerms that use
+        // the same property (for details see QueryBuilderData::uniqueVarName)
+        //
+        QList<Term> sortedSubTerms;
+        foreach( const Nepomuk::Query::Term &t, term.toAndTerm().subTerms() ) {
+            if(t.isComparisonTerm() && !t.toComparisonTerm().variableName().isEmpty()) {
+                sortedSubTerms.prepend(t);
+            }
+            else {
+                sortedSubTerms.append(t);
+            }
+        }
+
+        foreach(const Term& subTerm, sortedSubTerms) {
+            newAndTerm.addSubTerm(optimizeEvenMore(subTerm));
+        }
+        return newAndTerm;
+    }
+
+    //
+    // For OptionalTerms we simply need to optimize the subterm
+    //
+    else if(term.type() == Term::Optional) {
+        return OptionalTerm::optionalizeTerm(optimizeEvenMore(term.toOptionalTerm().subTerm()));
+    }
+
+    //
+    // For NegationTerms we simply need to optimize the subterm
+    //
+    else if(term.type() == Term::Negation) {
+        return NegationTerm::negateTerm(optimizeEvenMore(term.toNegationTerm().subTerm()));
+    }
+
     else {
-        return QString();
+        return term;
     }
 }
 
 
-
-QStringList Nepomuk::Query::QueryPrivate::buildRequestPropertyVariableList() const
-{
-    QStringList s;
-    for ( int i = 1; i <= m_requestProperties.count(); ++i ) {
-        s << QString( "?reqProp%1 " ).arg( i );
-    }
-    return s;
-}
-
-
-QString Nepomuk::Query::QueryPrivate::buildRequestPropertyPatterns() const
-{
-    QString s;
-    int i = 1;
-    foreach ( const Query::RequestProperty& rp, m_requestProperties ) {
-        if ( rp.optional() ) {
-            s += "OPTIONAL { ";
-        }
-
-        s += QString( "?r %1 ?reqProp%2 . " )
-             .arg( Soprano::Node::resourceToN3( rp.property().uri() ) )
-             .arg( i++ );
-
-        if ( rp.optional() ) {
-            s += "} ";
-        }
-    }
-    return s;
-}
-
-
-
 class Nepomuk::Query::Query::RequestProperty::Private : public QSharedData
 {
 public:
@@ -225,6 +284,12 @@ bool Nepomuk::Query::Query::isFileQuery() const
 }
 
 
+Nepomuk::Query::FileQuery Nepomuk::Query::Query::toFileQuery() const
+{
+    return FileQuery( *this );
+}
+
+
 Nepomuk::Query::Term Nepomuk::Query::Query::term() const
 {
     return d->m_term;
@@ -261,6 +326,42 @@ void Nepomuk::Query::Query::setOffset( int offset )
 }
 
 
+void Nepomuk::Query::Query::setFullTextScoringEnabled( bool enabled )
+{
+    d->m_fullTextScoringEnabled = enabled;
+}
+
+
+void Nepomuk::Query::Query::setFullTextScoringSortOrder( Qt::SortOrder order )
+{
+    d->m_fullTextScoringSortOrder = order;
+}
+
+
+bool Nepomuk::Query::Query::fullTextScoringEnabled() const
+{
+    return d->m_fullTextScoringEnabled;
+}
+
+
+Qt::SortOrder Nepomuk::Query::Query::fullTextScoringSortOrder() const
+{
+    return d->m_fullTextScoringSortOrder;
+}
+
+
+void Nepomuk::Query::Query::setQueryFlags( QueryFlags flags )
+{
+    d->m_flags = flags;
+}
+
+
+Nepomuk::Query::Query::QueryFlags Nepomuk::Query::Query::queryFlags() const
+{
+    return d->m_flags;
+}
+
+
 void Nepomuk::Query::Query::addRequestProperty( const RequestProperty& property )
 {
     d->m_requestProperties.append( property );
@@ -279,66 +380,90 @@ QList<Nepomuk::Query::Query::RequestProperty> Nepomuk::Query::Query::requestProp
 }
 
 
-namespace {
-    template<typename T>
-    bool compareQList( const QList<T>& rp1, const QList<T>& rp2 ) {
-        // brute force
-        foreach( const T& rp, rp1 ) {
-            if ( !rp2.contains( rp ) ) {
-                return false;
-            }
-        }
-        foreach( const T& rp, rp2 ) {
-            if ( !rp1.contains( rp ) ) {
-                return false;
-            }
-        }
-        return true;
-    }
-}
-
 bool Nepomuk::Query::Query::operator==( const Query& other ) const
 {
     return( d->m_limit == other.d->m_limit &&
             d->m_offset == other.d->m_offset &&
             d->m_term == other.d->m_term &&
             compareQList( d->m_requestProperties, other.d->m_requestProperties ) &&
-            compareQList( d->m_includeFolders, other.d->m_includeFolders ) &&
+            compareHash( d->m_includeFolders, other.d->m_includeFolders ) &&
             compareQList( d->m_excludeFolders, other.d->m_excludeFolders ) &&
             d->m_isFileQuery == other.d->m_isFileQuery &&
             d->m_fileMode == other.d->m_fileMode );
 }
 
 
-QString Nepomuk::Query::Query::toSparqlQuery( SparqlFlags flags ) const
+bool Nepomuk::Query::Query::operator!=( const Query& other ) const
+{
+    return( d->m_limit != other.d->m_limit ||
+            d->m_offset != other.d->m_offset ||
+            d->m_term != other.d->m_term ||
+            !compareQList( d->m_requestProperties, other.d->m_requestProperties ) ||
+            !compareHash( d->m_includeFolders, other.d->m_includeFolders ) ||
+            !compareQList( d->m_excludeFolders, other.d->m_excludeFolders ) ||
+            d->m_isFileQuery != other.d->m_isFileQuery ||
+            d->m_fileMode != other.d->m_fileMode );
+}
+
+
+QString Nepomuk::Query::Query::toSparqlQuery( SparqlFlags sparqlFlags ) const
 {
     Term term = d->m_term;
 
     // restrict to files if we are a file query
     if( d->m_isFileQuery ) {
-        //
-        // we do not use ResourceTypeTerm since we do not want to use crappy inference every time. All files have nfo:FileDataObject type anyway
-        //
         Term fileModeTerm;
-        ComparisonTerm fileTerm( Soprano::Vocabulary::RDF::type(), ResourceTerm(Vocabulary::NFO::FileDataObject()), ComparisonTerm::Equal );
-        ComparisonTerm folderTerm( Soprano::Vocabulary::RDF::type(), ResourceTerm(Vocabulary::NFO::Folder()), ComparisonTerm::Equal );
+        ResourceTypeTerm fileTerm( Vocabulary::NFO::FileDataObject() );
+        ResourceTypeTerm folderTerm( Vocabulary::NFO::Folder() );
         if( d->m_fileMode == FileQuery::QueryFiles )
             fileModeTerm = AndTerm( fileTerm, NegationTerm::negateTerm( folderTerm ) );
         else if( d->m_fileMode == FileQuery::QueryFolders )
             fileModeTerm = AndTerm( folderTerm, NegationTerm::negateTerm( fileTerm ) );
         else
             fileModeTerm = OrTerm( fileTerm, folderTerm );
-        term = AndTerm( term, fileModeTerm );
+        term = AndTerm( term, fileModeTerm, d->createFolderFilter() );
     }
 
+
+    // convert request properties into ComparisonTerms
+    // in ask and count query mode we can omit the optional req props
+    for ( int i = 0; i < d->m_requestProperties.count(); ++i ) {
+        const RequestProperty& rp = d->m_requestProperties[i];
+        ComparisonTerm rpt( rp.property(), Term() );
+        rpt.setVariableName( QString::fromLatin1("reqProp%1").arg(i+1) );
+        if( rp.optional() && !( sparqlFlags&(CreateAskQuery|CreateCountQuery) ) ) {
+            term = term && OptionalTerm::optionalizeTerm( rpt );
+        }
+        else if( !rp.optional() ) {
+            term = term && rpt;
+        }
+    }
+
+
     // optimize whatever we can
-    term = QueryPrivate::optimizeTerm( term );
+    term = term.optimized();
+
+    // perform internal optimizations
+    term = d->optimizeEvenMore(term);
 
     // actually build the SPARQL query patterns
-    QueryBuilderData qbd( flags );
+    QueryBuilderData qbd( d.constData(), sparqlFlags );
+
+
+    //
+    // We restrict results to user visible types only. There is no need for this with file queries as normally all files are visible.
+    //
+    QString userVisibilityRestriction;
+    if( !(queryFlags()&NoResultRestrictions) && !d->m_isFileQuery ) {
+        userVisibilityRestriction = QString::fromLatin1("FILTER EXISTS { ?r a [ %1 %2 ] . } . ")
+                                    .arg(Soprano::Node::resourceToN3(Soprano::Vocabulary::NAO::userVisible()),
+                                         Soprano::Node::literalToN3(Soprano::LiteralValue(true)));
+    }
+
+
     QString termGraphPattern;
     if( term.isValid() ) {
-        termGraphPattern = term.d_ptr->toSparqlGraphPattern( QLatin1String( "?r" ), &qbd );
+        termGraphPattern = term.d_ptr->toSparqlGraphPattern( QLatin1String( "?r" ), 0, userVisibilityRestriction, &qbd );
         if( termGraphPattern.isEmpty() ) {
             kDebug() << "Got no valid SPARQL pattern from" << term;
             return QString();
@@ -346,57 +471,59 @@ QString Nepomuk::Query::Query::toSparqlQuery( SparqlFlags flags ) const
     }
 
     // build the list of variables to select (in addition to the main result variable ?r)
-    QStringList selectVariables = d->buildRequestPropertyVariableList() + qbd.customVariables();
-    const QString scoringExpression = qbd.buildScoringExpression();
-    if( !scoringExpression.isEmpty() )
-        selectVariables << scoringExpression;
+    QStringList selectVariables = qbd.customVariables();
 
-    // restrict to resources from nrl:InstanceBases only. There is no need to do that for file queries
-    // as those alreday restrict the type. While in theory there might be file or folder resources in
-    // non-InstanceBase graphs it is very unlikely and omitting the instanceBaseRestriction will give
-    // us a performance gain - even if small it is worth it.
-    const QString instanceBaseRestriction =
-        QString::fromLatin1( "graph %1 { ?r a %2 . } . { %1 a %3 . } UNION { %1 a %4 . } . " )
-        .arg( qbd.uniqueVarName(),
-              qbd.uniqueVarName(),
-              Soprano::Node::resourceToN3(Soprano::Vocabulary::NRL::InstanceBase()),
-              Soprano::Node::resourceToN3(Soprano::Vocabulary::NRL::DiscardableInstanceBase()) );
+    // add additional scoring variable if requested
+    if( d->m_fullTextScoringEnabled ) {
+        const QString scoringExpression = qbd.buildScoringExpression();
+        if( !scoringExpression.isEmpty() )
+            selectVariables << scoringExpression;
+    }
 
     // build the core of the query - the part that never changes
-    QString queryBase = QString::fromLatin1( "%1 where { %2 %3 %4 %5 }" )
-                        .arg( selectVariables.join( QLatin1String(" " ) ),
-                              termGraphPattern,
-                              d->createFolderFilter( QLatin1String( "?r" ), &qbd ),
-                              d->buildRequestPropertyPatterns(),
-                              d->m_isFileQuery ? QString() : instanceBaseRestriction );
-
-    // add optional order terms
-    if( !(flags & CreateCountQuery) )
-        queryBase += qbd.buildOrderString();
-
-    // offset and limit
-    if ( d->m_offset > 0 )
-        queryBase += QString::fromLatin1( " OFFSET %1" ).arg( d->m_offset );
-    if ( d->m_limit > 0 )
-        queryBase += QString::fromLatin1( " LIMIT %1" ).arg( d->m_limit );
+    QString queryBase = QString::fromLatin1( "where { %1 }" )
+                        .arg( termGraphPattern );
 
     // build the final query
     QString query;
-    if( flags & CreateCountQuery ) {
+    if( sparqlFlags & CreateCountQuery ) {
         if( selectVariables.isEmpty() ) {
             // when there are no additional variables we can perfectly use count(distinct)
-            query = QLatin1String("select count(distinct ?r) as ?cnt ") + queryBase;
+            query = QString::fromLatin1("select count(distinct ?r) as ?cnt %1 %2")
+                    .arg( selectVariables.join( QLatin1String(" " ) ),
+                          queryBase );
         }
         else {
             // when there are additional variables we need to do some magic to count the
             // number of rows instead of a list of counts
-            query = QString::fromLatin1("select count(%1) as ?cnt where { { select count(*) as %1 ?r %2 } }")
-                    .arg(qbd.uniqueVarName(), queryBase );
+            // we cannot simply leave out the additional variables since that would change
+            // the number of results.
+            query = QString::fromLatin1("select count(%1) as ?cnt %2 where { { select count(*) as %1 ?r %3 } }")
+                    .arg(qbd.uniqueVarName(),
+                         selectVariables.join( QLatin1String(" " ) ),
+                         queryBase );
         }
     }
-    else {
-        query = QLatin1String( "select distinct ?r " ) + queryBase;
+    else if( sparqlFlags & CreateAskQuery ) {
+        query = QLatin1String( "ask ") + queryBase;
     }
+    else {
+        QString fullTextExcerptExpression;
+        if( !(queryFlags()&WithoutFullTextExcerpt) ) {
+            fullTextExcerptExpression = qbd.buildSearchExcerptExpression();
+        }
+        query = QString::fromLatin1( "select distinct ?r %1 %2 %3" )
+                .arg( selectVariables.join( QLatin1String(" " ) ),
+                      fullTextExcerptExpression,
+                      queryBase );
+        query += qbd.buildOrderString();
+    }
+
+    // offset and limit
+    if ( d->m_offset > 0 )
+        query += QString::fromLatin1( " OFFSET %1" ).arg( d->m_offset );
+    if ( d->m_limit > 0 )
+        query += QString::fromLatin1( " LIMIT %1" ).arg( d->m_limit );
 
     return query.simplified();
 }
@@ -414,14 +541,31 @@ KUrl Nepomuk::Query::Query::toSearchUrl( SparqlFlags flags ) const
 // would make the URL handling in the kio slave more complicated.... oh, well.
 KUrl Nepomuk::Query::Query::toSearchUrl( const QString& customTitle, SparqlFlags flags ) const
 {
+    // the nepomuksearch:/ KIO slave does not handle count or ask queries
     flags &= ~CreateCountQuery;
+    flags &= ~CreateAskQuery;
+
+    Query q( *this );
+
+    // the nepomuksearch:/ KIO slave does not make use of full text scores. Thus, we avoid the
+    // overhead by disabling them
+    q.setFullTextScoringEnabled( false );
+
     KUrl url( QLatin1String("nepomuksearch:/") );
     if( flags == NoFlags )
-        url.addQueryItem( QLatin1String("encodedquery"), toString() );
+        url.addQueryItem( QLatin1String("encodedquery"), q.toString() );
     else
-        url.addQueryItem( QLatin1String("sparql"), toSparqlQuery( flags ) );
-    if( !customTitle.isEmpty() )
-        url.addQueryItem( QLatin1String("title"), customTitle );
+        url.addQueryItem( QLatin1String("sparql"), q.toSparqlQuery( flags ) );
+
+    QString title(customTitle);
+    if( customTitle.isEmpty() ) {
+        title = titleFromQueryUrl( url );
+    }
+
+    // replace slashes with the "fraction slash" which is the same KIO::encodeFileName does.
+    // however, we do not want to link to KIO.
+    url.addPath( title.replace('/', QChar(0x2044)) );
+
     return url;
 }
 
@@ -435,12 +579,21 @@ Nepomuk::Query::RequestPropertyMap Nepomuk::Query::Query::requestPropertyMap() c
     return rpm;
 }
 
+
 QString Nepomuk::Query::Query::toString() const
 {
     return Nepomuk::Query::serializeQuery( *this );
 }
 
 
+Nepomuk::Query::Query Nepomuk::Query::Query::optimized() const
+{
+    Query newQuery( *this );
+    newQuery.setTerm( term().optimized() );
+    return newQuery;
+}
+
+
 // static
 Nepomuk::Query::Query Nepomuk::Query::Query::fromString( const QString& queryString )
 {
@@ -456,8 +609,7 @@ namespace {
         if( url.queryItems().contains( "query" ) ) {
             return url.queryItem( "query" );
         }
-        else if ( !url.queryItems().contains( "sparql" ) &&
-                  !url.queryItems().contains( "encodedquery" ) ) {
+        else if ( !url.hasQuery() ) {
             return url.path().section( '/', 0, 0, QString::SectionSkipEmpty );
         }
         else {
@@ -521,28 +673,51 @@ QString Nepomuk::Query::Query::titleFromQueryUrl( const KUrl& url )
         return QString();
     }
 
-    if ( url.hasQueryItem( QLatin1String( "title" ) ) ) {
-        return url.queryItem(QLatin1String( "title" ) );
-    }
-
-    QString userQuery( url.queryItem(QLatin1String( "userquery" ) ) );
-
-    if ( userQuery.isEmpty() ) {
-        userQuery = extractPlainQuery( url );
+    // the title is the first section of the path, but only if we have a query
+    if( url.hasQuery() ) {
+        QString title = url.path().section( '/', 0, 0, QString::SectionSkipEmpty );
+        if(!title.isEmpty())
+            return title;
     }
 
+    // no title in the path, try to get the user query, i.e. a non-encoded and non-sparql query
+    QString userQuery = extractPlainQuery( url );
     if ( !userQuery.isEmpty() ) {
         return i18nc( "@title UDS_DISPLAY_NAME for a KIO directory listing. %1 is the query the user entered.",
                       "Query Results from '%1'",
                       userQuery );
     }
 
-    // fallback
+    // fallback: generic query title
     return i18nc( "@title UDS_DISPLAY_NAME for a KIO directory listing.",
                   "Query Results");
 }
 
 
+Nepomuk::Query::Query Nepomuk::Query::operator&&( const Query& query, const Term& term )
+{
+    Query newQuery( query );
+    newQuery.setTerm( query.term() && term );
+    return newQuery;
+}
+
+
+Nepomuk::Query::Query Nepomuk::Query::operator||( const Query& query, const Term& term )
+{
+    Query newQuery( query );
+    newQuery.setTerm( query.term() || term );
+    return newQuery;
+}
+
+
+Nepomuk::Query::Query Nepomuk::Query::operator!( const Query& query )
+{
+    Query newQuery( query );
+    newQuery.setTerm( !query.term() );
+    return newQuery;
+}
+
+
 QDebug operator<<( QDebug dbg, const Nepomuk::Query::Query& query )
 {
     return dbg << query.toString();
